﻿using System;
using System.Drawing;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

namespace XImage_Class
{
	/// <summary> Create a bitmap crossed with an integer array.
	/// The array can be directly modified without recreating the bitmap.
	/// The array is pinned in memory reducing the efficiency of garbage collection.
	/// The class simply ignores attempts to write outside the array/bitmap.
	/// Read returns Color.Black which never equals comparison to a normal value.
	/// </summary>
	public class XImage
	{
		#region Class data

		GCHandle _handle;
		IntPtr _pointer = (IntPtr) null;
		Bitmap _bitmap = null;
		int[] _pixels = null;
		int _width = 0;
		int _height = 0;
		// raw length of the allocated bitmap that may be greater than _width * _height
		int _length = 0;

		#endregion
		#region  Properties
		/// <summary> A bitmap that can be directly modified. </summary>
		public Bitmap Bitmap
		{
			get { return _bitmap; }
			set { AllocateBitmap (value); }
		}
		/// <summary> Width of current <see cref="Bitmap"/>. </summary>
		public int Width
		{
			get { return _width; }
		}
		/// <summary> Height of current <see cref="Bitmap"/>. </summary>
		public int Height
		{
			get { return _height; }
		}
		/// <summary> Length of current array of <see cref="Bitmap"/>. </summary>
		public int Length
		{
			get { return _width*_height; }
		}
		/// <summary> The integer array that makes up the pixels in <see cref="Bitmap"/>.
		/// The pixels can be modified directly by modifying this array.
		/// The pixels must be from Color.FromArgb(a, r, g, b).ToArgb () or
		/// similar and accessed by premultiplied x and y index values
		/// </summary>
		public int[] Array
		{
			get { return _pixels; }
		}
		#endregion
		#region Constructors
		/// <summary> Allocate a new blank transparent bitmap
		/// </summary>
		/// <param name="x">Width</param>
		/// <param name="y">Height</param>
		public XImage (int x, int y)
		{
			AllocateBitmap (x, y);
		}
		/// <summary> Allocate a new bitmap copy
		/// </summary>
		/// <param name="bitmap">Bitmap to allocate and copy</param>
		public XImage (Bitmap a)
		{
			AllocateBitmap (a);
		}
		#endregion
		#region Methods
		/// <summary> Reallocate a new bitmap copy.
		/// </summary>
		/// <param name="bitmap">Bitmap to copy</param>
		public void AllocateBitmap (Bitmap a)
		{
			RawAllocateBitmap (a.Width, a.Height);
			for (int x = 0; x < _width; ++x)
				for (int y = 0; y < _height; ++y)
					SetPixel (x, y, a.GetPixel (x, y));
		}
		/// <summary> Reallocate a new blank transparent bitmap
		/// </summary>
		/// <param name="width">Width</param>
		/// <param name="height">Height</param>
		public void AllocateBitmap (int width, int height)
		{
			RawAllocateBitmap (Width, Height);
			for (int i = 0; i < _width * _height; ++i)
				_pixels[i] = 0;
		}
		/// <summary> Reallocate a new uninitialized bitmap
		/// </summary>
		/// <param name="width">Width</param>
		/// <param name="height">Height</param>
		void RawAllocateBitmap (int width, int height)
		{
			if (0 >= width || 0 >= height)
				width = height = 0;
			// no changes
			if (_width == width && _height == height)
				return;
			_width = width;
			_height = height;
			// forced unallocate!
			if (0 == _width)
			{
				Dispose ();
				return;
			}
			// do we already have sufficient space?
			if (_length < width * height)
			{
				Dispose ();			
				// 4 bytes per pixel, format: ARGB
				_length = width * height;
				_pixels = new int[_length];
				// if not pinned the GC can move the array around
				_handle = GCHandle.Alloc (_pixels, GCHandleType.Pinned);
				_pointer = Marshal.UnsafeAddrOfPinnedArrayElement (_pixels, 0);
			}
			_bitmap = new Bitmap (width, height, width * 4, PixelFormat.Format32bppArgb, _pointer);
		}
		public void Dispose ()
		{
			if (_bitmap != null)
			{
				_handle.Free ();
				_bitmap.Dispose ();
				_pointer = (IntPtr) null;
				_bitmap = null;
				_pixels = null;
				_width = 0;
				_height = 0;
				_length = 0;
			}
		}
		/// <summary> Get pixel in the linear array.
		/// Return Color.Black for any access outside the array.
		/// </summary>
		/// <param name="i">I</param>
		public Color GetPixel (int i)
		{
			if (i < 0 || i >= _width * _height)
				return Color.Black;
			return Color.FromArgb (_pixels[i]);
		}
		/// <summary> Get pixel in the bitmap.
		/// Return Color.Black for any access outside the bitmap.
		/// </summary>
		/// <param name="x">X</param>
		/// <param name="x">Y</param>
		public Color GetPixel (int x, int y)
		{
			if (x < 0 || y < 0 || x >= _width || y >= _height)
				return Color.Black;
			return Color.FromArgb (_pixels[x + y * _width]);
		}
		/// <summary> Set pixel in the linear array.
		/// Ignore any access outside the array.
		/// </summary>
		/// <param name="i">I</param>
		/// <param name="argb">argb Color value</param>
		public void SetPixel (int i, Color argb)
		{
			if (i < 0 || i >= _width * _height)
				return;
			_pixels[i] = argb.ToArgb ();
		}
		/// <summary> Set pixel in the bitmap.
		/// Ignore any access outside the bitmap.
		/// </summary>
		/// <param name="x">X</param>
		/// <param name="x">Y</param>
		/// <param name="argb">argb Color value</param>
		public void SetPixel (int x, int y, Color argb)
		{
			if (x < 0 || y < 0 || x >= _width || y >= _height)
				return;
			_pixels[x + y * _width] = argb.ToArgb ();
		}
		#endregion
	}
}
